/*********************************************************************
*                            K E R C O N                             *
**********************************************************************
* Modified version of Situs vol2pdb.c (c) Willy Wriggers, 1998-2001  *
* Specialized tool for kernel convolution                            *
**********************************************************************
*                                                                    * 
* compile on Unix platforms:                                         *
*               cc -O -c stdread.c                                   *
*               cc -O -c volio.c                                     *
*               cc -O -c pdbio.c                                     *
*               cc -O -c kercon.c                                    *
*               cc -O -o kercon stdread.o pdbio.o kercon.o volio.o   *
*                                                                    *
**********************************************************************
* See legal statement for terms of distribution                      *
*********************************************************************/

#include "situs.h"

/* global variables */

static int extx, exty;    /* kernel map */
static int extx2, exty2;  /* protein map */
static int extx3, exty3;  /* output map */

/* function declarations */

extern void readpdb(char *, int *, PDB **);
extern void writevol(char *, MYFLT, MYFLT, MYFLT, MYFLT, int, int, int, MYFLT *);
extern int readlnint ();
extern MYFLT readlnmyflt ();
void CopyAtoms (PDB *, PDB *, int, int, int);
char *Element (int);
int IsIgnored (PDB *, int);
unsigned long g1idz(int,int,int);
unsigned long g2idz(int,int,int); 
unsigned long g3idz(int,int,int); 

main(int argc, char *argv[])

{ MYFLT *phi, *phi2, *phi3;
  int extz, extz2, extz3, exth, indx, indy, indz, i, j, k;
  unsigned long nvox, nvox2, nvox3, count, indv, extxy2;
  MYFLT gridx2, gridy2, gridz2, gridx3, gridy3, gridz3, width, dval, dsqu;
  int x0, y0, z0, x1, y1, z1;
  MYFLT gx, gy, gz, a, b, c;
  MYFLT minx, miny, minz, maxx, maxy, maxz;
  MYFLT maxdist, currdist, currdistsq, kmsd, reso;
  int menumode, wmode, corrmode, margin;
  MYFLT bvalue, cvalue;
  MYFLT kampl, champl, sigmap, varmap, varp;
  MYFLT xp1, xp2, yp1, yp2, zp1, zp2, sqrad1, sqrad2, sqrtrad1, sqrtrad2;
  MYFLT sig1, rh1, rc1;
  int numA, numS, numU;
  PDB *pdbA, *pdbS;

  if (argc != 3) {
    fprintf(stderr, "kercon> Usage: kercon inputfile (PDB) outputfile (map)\n");
    exit (1);
  } 
  
  readpdb(argv[1], &numA, &pdbA);

  printf ("kercon> What kind of 3D density function do you want to create:\n");  
  printf ("kercon> \n"); 
  printf ("kercon>      1: Charge density (atom charges will be read from PDB occupancy field)\n");
  printf ("kercon>      2: Mass density (atom masses are assigned automatically)\n");
  printf ("kercon> ");
  wmode = readlnint();

  switch (wmode) {
  case 1:
    readpdb(argv[1], &numS, &pdbS);
    break;
  case 2:
    /* create the mass-weighted atom dataset */
    pdbS = (PDB *) malloc((MAXPDB)*sizeof(PDB));
    if (pdbS== NULL) {
      fprintf(stderr, "kercon> Error: Could not satisfy memory allocation request [e.c. 80005]\n"); 
      exit(80005);
    }
    numS = 0;
    numU = 0;
    for (i=0;i<numA;++i) {
      if (numS>=MAXPDB) {
	fprintf(stderr, "kercon> Error: Internal number of atoms exceed MAXPDB [e.c. 80010]\n"); 
	exit(80010);
      }
      if (IsIgnored(pdbA,i)) continue; 
      for (j=2;j<30;++j) if (strcasecmp(pdbA[i].atomType,Element(j))==0) {
	if (j<3) count = 0;
	else if (j<10) count = 1;
	else if (j<16) count = 2;
	else if (j<22) count = 3;
	else if (j<28) count = 4;
	else count = 5;	
	break;
      }
      
      if (j==30) {
	count = 1; /* default if element not recognized */
      }
      
      CopyAtoms (pdbA, pdbS, i, numS, count);
      numS += count; ++numU;
    }
    
    printf ("kercon> There are %d non-hydrogen atoms, represented by %d equally weighted input atoms \n", numU, numS);
    printf ("kercon> \n");
    break;
  default:
    fprintf(stderr, "kercon> Error: Could not identify option [e.c. 80020]\n"); 
    exit(80020);
  }    
  free(pdbA);
  
  /* measure protein extent and read voxel spacing */
  
  minx = 1e20; miny = 1e20; minz = 1e20; maxx = -1e20; maxy = -1e20; maxz = -1e20;
  for (i=0;i<numS;++i) {
    if (minx > pdbS[i].x) minx = pdbS[i].x;
    if (maxx < pdbS[i].x) maxx = pdbS[i].x;
    if (miny > pdbS[i].y) miny = pdbS[i].y;
    if (maxy < pdbS[i].y) maxy = pdbS[i].y;
    if (minz > pdbS[i].z) minz = pdbS[i].z;
    if (maxz < pdbS[i].z) maxz = pdbS[i].z;
  }
  printf ("kercon> The input structure measures %6.3f x %6.3f x %6.3f Angstrom\n",maxx-minx,maxy-miny,maxz-minz);
  printf ("kercon> \n");
  printf ("kercon> Please enter the desired voxel spacing for the output map (in Angstrom): ");
  width = readlnmyflt();
  if (width < 0.1) {
    printf ("kercon> Voxel spacing set to minimum value: 0.1 Angstrom\n");
    width = 0.1;
  } 
  printf ("kercon> \n");

  /* bring lattice into register with origin */
  minx = width*floor(minx/width);
  maxx = width*ceil(maxx/width);
  miny = width*floor(miny/width);
  maxy = width*ceil(maxy/width);
  minz = width*floor(minz/width);
  maxz = width*ceil(maxz/width);

  /* allocate protein density map */
  margin = 2;
  extx2 = ceil((maxx-minx)/width) + 2*margin + 1;
  exty2 = ceil((maxy-miny)/width) + 2*margin + 1;
  extz2 = ceil((maxz-minz)/width) + 2*margin + 1;
  gridx2 = minx - margin * width;
  gridy2 = miny - margin * width;
  gridz2 = minz - margin * width;
  nvox2 = extx2*exty2*extz2;
  phi2 = (MYFLT *) malloc(nvox2 * sizeof(MYFLT));
  if (phi2 == NULL) {
    fprintf(stderr, "kercon> Error: Could not satisfy memory allocation request [e.c. 80110]\n"); 
    exit(80110);
  }

  printf ("kercon> Please select the type of kernel:\n");  
  printf ("kercon> \n"); 
  printf ("kercon>      1: Gaussian: A exp(-1.5 r^2 / sigma^2),\n");
  printf ("kercon>         useful for resolution-lowering of atomic structures. \n"); 
  printf ("kercon> \n");
  printf ("kercon>      2: Hard Sphere: 0 (outside) or A (inside), \n");
  printf ("kercon>         useful for bead-modeling at reduced complexity:\n");
  printf ("kercon>         - sphere radii are read from input PDB file (B-factor field).\n");
  printf ("kercon>         - sphere boundaries are anti-aliased 1-voxel wide\n");
  printf ("kercon> ");
  menumode = readlnint();
  if (menumode < 1 || menumode > 2) {
    fprintf(stderr, "kercon> Error: Could not identify option [e.c. 80120]\n"); 
    exit(80120);
  }

  if (menumode == 1) {
    printf ("kercon> Kernel width. Please enter (in Angstrom):\n");
    printf ("kercon>     (as pos. value) target resolution (== 2 sigma) or\n"); 
    printf ("kercon>     (as neg. value) kernel half-max radius \n");
    printf ("kercon> Now enter (signed) value: ");
    reso = readlnmyflt();
    if (reso > 0) {
      sig1 = reso/(2.0);
      rh1 = sig1 * sqrt(log(2.0)) / sqrt(1.5);
    } else {
      rh1 = -reso;
      sig1 = rh1 * sqrt(1.5) / sqrt(log(2.0));
    }
    rc1 = sqrt(3.0)*sig1;
  }

  if (menumode == 1 && rh1/width >= 1.0) {
    printf ("kercon> \n");
    printf ("kercon> The Gaussian kernel has the following properties:\n");  
    printf ("kercon> \n"); 
    printf ("kercon>     Gaussian, A exp(-1.5 r^2 / sigma^2)\n");
    printf ("kercon>     sigma = %6.3fA, r-half = %6.3fA, r-cut = %6.3fA\n", sig1, rh1, rc1); 
    printf ("kercon> \n"); 
    printf ("kercon> Please enter the desired kernel amplitude A: ");
    kampl = readlnmyflt();
    printf ("kercon> \n");
  } else if (menumode == 1) {
    kampl = 1;
  }

  if (menumode == 2) for (i=0;i<numS;++i) if (pdbS[i].tempFactor <=0) {
    fprintf (stderr, "kercon> Error: All B-factor entries (= sphere radii) in file %s must be > 0 for this option [e.c. 80025]\n",argv[1]); 
    exit(80025);
  }
  
  if (menumode == 2) {
    printf ("kercon> \n"); 
    printf ("kercon> Please enter the desired amplitude A for the largest sphere");
    if (wmode == 1) printf ("\nkercon> (amplitudes of smaller spheres will be > A to maintain kernel volume and charge balance): ");
    else printf (": ");
    kampl = readlnmyflt();
    printf ("kercon> \n");
  }

  if (menumode == 1 && rh1/width >= 1.0) {
    printf ("kercon> Do you want to correct sigma for spreading introduced by tri-linear projection to lattice?\n");
    printf ("kercon> \n"); 
    printf ("kercon>      1: Yes (slightly lowers the kernel width to maintain target resolution) \n"); 
    printf ("kercon>      2: No \n"); 
    printf ("kercon> "); 
    corrmode = readlnint();
    if (corrmode < 1 || corrmode > 2) {
      fprintf(stderr, "kercon> Error: Could not identify option [e.c. 80125]\n"); 
      exit(80125);
    }
  } else corrmode = 2;


  if (menumode == 1) { /* interpolate structure to protein map and keep track of variability */
    for (count=0;count<nvox2;count++) *(phi2+count) = 0;
    varp = 0;
    if (wmode == 2) {
      printf ("kercon> Projecting atom masses to cubic lattice by tri-linear interpolation... \n");  
      for (i=0;i<numS;++i) {
	/* compute position within grid  */
	gx = margin + (pdbS[i].x - minx) / width;
	gy = margin + (pdbS[i].y - miny) / width;
	gz = margin + (pdbS[i].z - minz) / width;
	x0 = floor (gx); 
	y0 = floor (gy); 
	z0 = floor (gz); 
	x1 = x0+1;
	y1 = y0+1; 
	z1 = z0+1; 
	/* interpolate */
	a = x1-gx;
	b = y1-gy;
	c = z1-gz;
	*(phi2+g2idz(z0,y0,x0)) += a * b * c; varp += a * b * c * ((1-a)*(1-a)+(1-b)*(1-b)+(1-c)*(1-c));
	*(phi2+g2idz(z1,y0,x0)) += a * b * (1-c); varp += a * b * (1-c) * ((1-a)*(1-a)+(1-b)*(1-b)+c*c);
	*(phi2+g2idz(z0,y1,x0)) += a * (1-b) * c; varp += a * (1-b) * c * ((1-a)*(1-a)+b*b+(1-c)*(1-c));
	*(phi2+g2idz(z0,y0,x1)) += (1-a) * b * c; varp += (1-a) * b * c * (a*a+(1-b)*(1-b)+(1-c)*(1-c));
	*(phi2+g2idz(z1,y1,x0)) += a * (1-b) * (1-c); varp += a * (1-b) * (1-c) * ((1-a)*(1-a)+b*b+c*c);
	*(phi2+g2idz(z0,y1,x1)) += (1-a) * (1-b) * c; varp += (1-a) * (1-b) * c * (a*a+b*b+(1-c)*(1-c));
	*(phi2+g2idz(z1,y0,x1)) += (1-a) * b * (1-c); varp += (1-a) * b * (1-c) * (a*a+(1-b)*(1-b)+c*c);
	*(phi2+g2idz(z1,y1,x1)) += (1-a) * (1-b) * (1-c); varp += (1-a) * (1-b) * (1-c) * (a*a+b*b+c*c);
      }
    } else {
      printf ("kercon> Projecting charges to cubic lattice by tri-linear interpolation... \n");  
      for (i=0;i<numS;++i) {
	/* compute position within grid  */
	gx = margin + (pdbS[i].x - minx) / width;
	gy = margin + (pdbS[i].y - miny) / width;
	gz = margin + (pdbS[i].z - minz) / width;
	x0 = floor (gx); 
	y0 = floor (gy); 
	z0 = floor (gz); 
	x1 = x0+1;
	y1 = y0+1; 
	z1 = z0+1; 
	/* interpolate */
	a = x1-gx;
	b = y1-gy;
	c = z1-gz;
	*(phi2+g2idz(z0,y0,x0)) += pdbS[i].occupancy * a * b * c; varp += a * b * c * ((1-a)*(1-a)+(1-b)*(1-b)+(1-c)*(1-c));
	*(phi2+g2idz(z1,y0,x0)) += pdbS[i].occupancy * a * b * (1-c); varp += a * b * (1-c) * ((1-a)*(1-a)+(1-b)*(1-b)+c*c);
	*(phi2+g2idz(z0,y1,x0)) += pdbS[i].occupancy * a * (1-b) * c; varp += a * (1-b) * c * ((1-a)*(1-a)+b*b+(1-c)*(1-c));
	*(phi2+g2idz(z0,y0,x1)) += pdbS[i].occupancy * (1-a) * b * c; varp += (1-a) * b * c * (a*a+(1-b)*(1-b)+(1-c)*(1-c));
	*(phi2+g2idz(z1,y1,x0)) += pdbS[i].occupancy * a * (1-b) * (1-c); varp += a * (1-b) * (1-c) * ((1-a)*(1-a)+b*b+c*c);
	*(phi2+g2idz(z0,y1,x1)) += pdbS[i].occupancy * (1-a) * (1-b) * c; varp += (1-a) * (1-b) * c * (a*a+b*b+(1-c)*(1-c));
	*(phi2+g2idz(z1,y0,x1)) += pdbS[i].occupancy * (1-a) * b * (1-c); varp += (1-a) * b * (1-c) * (a*a+(1-b)*(1-b)+c*c);
	*(phi2+g2idz(z1,y1,x1)) += pdbS[i].occupancy * (1-a) * (1-b) * (1-c); varp += (1-a) * (1-b) * (1-c) * (a*a+b*b+c*c);
      }
    }
    varp /= ((double) numS);
    printf ("kercon> ... done. Lattice spread (rmsd): %6.3f Angstrom\n",width*sqrt(varp)); 
  }

  if (menumode == 1 && rh1/width >= 1.0) { /* compute lattice-noise corrected kernel map */
    printf ("kercon> \n");
    if (corrmode == 2)  printf ("kercon> Computing Gaussian kernel (no lattice correction) ... \n");
    else printf ("kercon> Computing Gaussian kernel (correcting sigma for lattice smoothing)... \n");
    kmsd = sig1*sig1/(width*width);
    reso = 2*sig1; if (corrmode == 2) reso = 2*sqrt(sig1*sig1+varp*width*width); /* variances are additive for uncorrelated samples */
    varmap = kmsd; if (corrmode == 1) varmap -= varp; /* variances are additive for uncorrelated samples */
    if (varmap < 0) {
      fprintf(stderr, "kercon> Error: lattice smoothing exceeds kernel size [e.c. 80130]\n"); 
      exit(80130);
    }
    sigmap = sqrt(varmap/3.0); /* sigma-1D */
    exth = (int) ceil(3*sigmap); /* truncate at 3 sigma-1D == sqrt(3) sigma-3D */
    extx = 2 * exth + 1; exty = extx; extz = extx; nvox = extx * exty * extz;
    phi = (MYFLT *) malloc(nvox * sizeof(MYFLT));
    if (phi == NULL) {
      fprintf(stderr, "kercon> Error: Could not satisfy memory allocation request [e.c. 80140]\n"); 
      exit(80140);
    }
    /* write Gaussian within 3 sigma-1D to map */
    bvalue = -1 / (2.0*sigmap*sigmap);
    cvalue = 9*sigmap*sigmap;
    for(count=0;count<nvox;count++) *(phi+count) = 0;
    for (indz=0;indz<extz;indz++)
    for (indy=0;indy<exty;indy++)
    for (indx=0;indx<extx;indx++) {
      dsqu = (indx-exth)*(indx-exth)+(indy-exth)*(indy-exth)+(indz-exth)*(indz-exth);
      if (dsqu < cvalue) *(phi+g1idz(indz,indy,indx)) = kampl * exp (dsqu * bvalue);
    }
    printf ("kercon> ... done. Kernel map extent %d x %d x %d voxels \n", extx, exty, extz);
    printf ("kercon> \n");
  }
 
  /* convolve and write output */

  if (menumode == 1 && rh1/width >= 1.0) {
    printf ("kercon> Convolving lattice with kernel... \n");  
    /* allocate output density map */
    extx3 = ceil((maxx-minx)/width) + 2*exth + 2*margin + 1;
    exty3 = ceil((maxy-miny)/width) + 2*exth + 2*margin + 1;
    extz3 = ceil((maxz-minz)/width) + 2*exth + 2*margin + 1;
    gridx3 = minx - (exth+margin) * width;
    gridy3 = miny - (exth+margin) * width;
    gridz3 = minz - (exth+margin) * width;
    nvox3 = extx3*exty3*extz3;
    phi3 = (MYFLT *) malloc(nvox3 * sizeof(MYFLT));
    if (phi3 == NULL) {
      fprintf(stderr, "kercon> Error: Could not satisfy memory allocation request [e.c. 80310]\n"); 
      exit(80310);
    }
    for (count=0;count<nvox3;count++) *(phi3+count) = 0;
    extxy2 = extx2 * exty2;
    for (count=0;count<nvox2;count++) {
      dval = *(phi2+count);
      if (dval != 0) {
	indv = count;
	k = indv / extxy2;
	indv -= k * extxy2;
	j = indv / extx2;
	i = indv - j * extx2;
	for (indz=0;indz<extz;indz++) for (indy=0;indy<exty;indy++) for (indx=0;indx<extx;indx++) 
	  *(phi3+g3idz(k+indz,j+indy,i+indx)) += *(phi+g1idz(indz,indy,indx)) * dval;
      }
    }
    printf ("kercon> ... done. Spatial resolution (2 sigma) of output map: %6.3fA \n",reso);
    if (corrmode == 2)  printf ("kercon> (slightly larger than target resolution due to uncorrected lattice smoothing)\n");
    printf ("kercon> \n");
    writevol(argv[2], width, gridx3, gridy3, gridz3, extx3, exty3, extz3, phi3);
  } else if (menumode == 1) {
    writevol(argv[2], width, gridx2, gridy2, gridz2, extx2, exty2, extz2, phi2);
    printf ("kercon> \n");
    printf ("kercon> Warning: kernel width should be large relative to lattice voxel spacing! \n");
    printf ("kercon> Interpolated map written to file %s WITHOUT kernel convolution.\n", argv[2]);
    printf ("kercon> Spatial resolution (2 sigma) from lattice spread alone: %6.3fA\n",width*sqrt(varp)*2.0);
    printf ("kercon> If kernel convolution is desired, increase the kernel width or decrease the lattice spacing. \n");
  } else if (menumode == 2) {
    printf ("kercon> Convolving lattice with kernel... \n");  
    /* determine maximum radius */
    maxdist = 0;
    for (i=0;i<numS;++i) {
      currdist = pdbS[i].tempFactor;
      if (currdist > maxdist) maxdist = currdist;
    }      
    exth = ceil(maxdist/width);

    /* allocate output density map */
    extx3 = ceil((maxx-minx)/width) + 2*exth + 2*margin + 1;
    exty3 = ceil((maxy-miny)/width) + 2*exth + 2*margin + 1;
    extz3 = ceil((maxz-minz)/width) + 2*exth + 2*margin + 1;
    gridx3 = minx - (exth+margin) * width;
    gridy3 = miny - (exth+margin) * width;
    gridz3 = minz - (exth+margin) * width;
    nvox3 = extx3*exty3*extz3;
    phi3 = (MYFLT *) malloc(nvox3 * sizeof(MYFLT));
    if (phi3 == NULL) {
      fprintf(stderr, "kercon> Error: Could not satisfy memory allocation request [e.c. 80311]\n"); 
      exit(80311);
    }

    for (count=0;count<nvox3;count++) *(phi3+count) = 0;
    if (wmode==2) {
      for (i=0;i<numS;++i) {
	printf ("kercon> Now processing input atom nr. %d (out of %d)...\n",i+1,numS); 
	xp2 = pdbS[i].x;
	yp2 = pdbS[i].y;
	zp2 = pdbS[i].z;
	sqrtrad1 = pdbS[i].tempFactor-0.5*width; if (sqrtrad1<0) sqrtrad1=0;
	sqrad1 = sqrtrad1*sqrtrad1;
	sqrtrad2 = pdbS[i].tempFactor+0.5*width; if (sqrtrad2<0) sqrtrad2=0;
	sqrad2 = sqrtrad2*sqrtrad2;
	for (indz=0;indz<extz3;indz++) for (indy=0;indy<exty3;indy++) for (indx=0;indx<extx3;indx++) { 
	  xp1 = indx * width + gridx3;
	  yp1 = indy * width + gridy3;
	  zp1 = indz * width + gridz3;
	  currdistsq = (xp1-xp2)*(xp1-xp2)+(yp1-yp2)*(yp1-yp2)+(zp1-zp2)*(zp1-zp2);
	  if (currdistsq<sqrad1) *(phi3+g3idz(indz,indy,indx)) += kampl;
	  else if (currdistsq<sqrad2) {
	    currdist = sqrt(currdistsq);
	    *(phi3+g3idz(indz,indy,indx)) += kampl * (sqrtrad2-currdist) / (sqrtrad2-sqrtrad1);
	  }
	}
      }
    } else {
      for (i=0;i<numS;++i) {
	printf ("kercon> Now processing input atom nr. %d (out of %d)...\n",i+1,numS); 
	xp2 = pdbS[i].x;
	yp2 = pdbS[i].y;
	zp2 = pdbS[i].z;
	champl = kampl * pdbS[i].occupancy * pow(maxdist,3) / pow(pdbS[i].tempFactor,3);
	sqrtrad1 = pdbS[i].tempFactor-0.5*width; if (sqrtrad1<0) sqrtrad1=0;
	sqrad1 = sqrtrad1*sqrtrad1;
	sqrtrad2 = pdbS[i].tempFactor+0.5*width; if (sqrtrad2<0) sqrtrad2=0;
	sqrad2 = sqrtrad2*sqrtrad2;
	for (indz=0;indz<extz3;indz++) for (indy=0;indy<exty3;indy++) for (indx=0;indx<extx3;indx++) { 
	  xp1 = indx * width + gridx3;
	  yp1 = indy * width + gridy3;
	  zp1 = indz * width + gridz3;
	  currdistsq = (xp1-xp2)*(xp1-xp2)+(yp1-yp2)*(yp1-yp2)+(zp1-zp2)*(zp1-zp2);
	  if (currdistsq<sqrad1) *(phi3+g3idz(indz,indy,indx)) += champl;
	  else if (currdistsq<sqrad2) {
	    currdist = sqrt(currdistsq);
	    *(phi3+g3idz(indz,indy,indx)) += champl * (sqrtrad2-currdist) / (sqrtrad2-sqrtrad1);
	  }
	}
      }
    }
      
    printf ("kercon> ... done. Maximum kernel radius applied: %6.3fA\n",maxdist);
    printf ("kercon> \n");
    writevol(argv[2], width, gridx3, gridy3, gridz3, extx3, exty3, extz3, phi3);
  }
}

void CopyAtoms (PDB *pdbP, PDB *pdbS, int ip, int is, int ktot)
/* copies ktot duplicates of atom # pi of pdbP to pdbS starting at # si */  

{ int j, k;

  for (k=0;k<ktot;++k) {
    pdbS[is+k].x = pdbP[ip].x;   
    pdbS[is+k].y = pdbP[ip].y;  
    pdbS[is+k].z = pdbP[ip].z; 
    for (j=0;j<4;++j) pdbS[is+k].segID[j] = pdbP[ip].segID[j];
    pdbS[is+k].serial = is+k+1;	
    for (j=0;j<7;++j) pdbS[is+k].recdName[j] = pdbP[ip].recdName[j];
    for (j=0;j<3;++j) pdbS[is+k].atomType[j] = pdbP[ip].atomType[j];
    for (j=0;j<3;++j) pdbS[is+k].atomLoc[j] = pdbP[ip].atomLoc[j];
    for (j=0;j<2;++j) pdbS[is+k].altLoc[j] = pdbP[ip].altLoc[j];
    for (j=0;j<5;++j) pdbS[is+k].resName[j] = pdbP[ip].resName[j]; 
    for (j=0;j<2;++j) pdbS[is+k].chainID[j] = pdbP[ip].chainID[j];
    pdbS[is+k].resSeq = pdbP[ip].resSeq;
    for (j=0;j<2;++j) pdbS[is+k].iCode[j] = pdbP[ip].iCode[j]; 
    pdbS[is+k].occupancy = pdbP[ip].occupancy;   
    pdbS[is+k].tempFactor = pdbP[ip].tempFactor;
    pdbS[is+k].ftNote = pdbP[ip].ftNote;  
    for (j=0;j<3;++j) pdbS[is+k].element[j] = pdbP[ip].element[j];
    for (j=0;j<3;++j) pdbS[is+k].charge[j] = pdbP[ip].charge[j];
  }
  return;
}

char *Element (int n)
/* returns first 30 element symbols */

{ static char *name[] = {
  "H", "HE", "LI", "BE", "B", "C", "N", "O", "F", "NE", 
  "NA", "MG", "AL", "SI", "P", "S", "CL", "AR", "K", "CA", 
  "SC", "TI", "V", "CR", "MN", "FE", "CO", "NI", "CU", "ZN" };
  if ((n<0)||(n>=30)) {	
    fprintf(stderr, "kercon> Error: element index out of range [e.c. 80410]\n"); 
    exit(80410);
    return 0;
  } else return name[n];
}

int IsIgnored (PDB *pdbX, int i)
/* checks if i-th atom in pdbX is a hydrogen */
/* Note: false hydrogen positives for Hg, Hf, Ho (these are very rare in proteins) */
{
  if (pdbX[i].atomType[0] =='H' || pdbX[i].atomType[0] == 'h') return 1;
  else return 0;
}


unsigned long g1idz(int k, int j, int i) 
{ return extx*exty*k+extx*j+i; }

unsigned long g2idz(int k, int j, int i) 
{ return extx2*exty2*k+extx2*j+i; }

unsigned long g3idz(int k, int j, int i) 
{ return extx3*exty3*k+extx3*j+i; }



