/*********************************************************************
*                           P I N D O W N                            *
**********************************************************************
* Program is part of the Situs package (c) Willy Wriggers, 1998-2001 *
* URL: http://situs.scripps.edu                                      *
**********************************************************************
*                                                                    * 
* compile on Unix platforms:                                         *
*               cc -O -c stdread.c                                   *
*               cc -O -c volio.c                                     *
*               cc -O -c pindown.c                                   *
*               cc -O -o pindown pindown.o stdread.o volio.o         *
*                                                                    *
**********************************************************************
* See legal statement for terms of distribution                      *
*********************************************************************/

#include "situs.h"          

extern int readlnint ();
extern void readvol(char *, MYFLT *, MYFLT *, MYFLT *, MYFLT *, int *, int *, int *, MYFLT **);
extern void writevol(char *, MYFLT, MYFLT, MYFLT, MYFLT, int, int, int, MYFLT *);
unsigned long pindex(int,int,int);
  
static int extx, exty, extz;

main(int argc, char *argv[])

{ MYFLT *phi, *phi2;
  int indx, indy, indz;  
  unsigned long nvox, count;
  MYFLT gridx, gridy, gridz, width;
  int minx, maxx, miny, maxy, minz, maxz;


  if (argc != 3) {
    fprintf(stderr, "pindown> Usage: pindown inputfile outputfile \n");
    exit (1);
  }

  readvol(argv[1], &width, &gridx, &gridy, &gridz, &extx, &exty, &extz, &phi);
   
  printf ("pindown> Voxel range of new grid. Enter values:\n");
  printf ("pindown> Minimum x value (1-%d): ",extx);
  minx = readlnint()-1; if (minx<0) minx = 0; if (minx>=extx) minx = extx-1;
  printf ("pindown> Maximum x value (%d-%d): ",minx+1,extx);
  maxx = readlnint()-1; if (maxx<minx) maxx = minx; if (maxx>=extx) maxx = extx-1;
  printf ("pindown> Minimum y value (1-%d): ",exty);
  miny = readlnint()-1; if (miny<0) miny = 0; if (miny>=exty) miny = exty-1;
  printf ("pindown> Maximum y value (%d-%d): ",miny+1,exty);
  maxy = readlnint()-1; if (maxy<miny) maxy = miny; if (maxy>=exty) maxy = exty-1;
  printf ("pindown> Minimum z value (1-%d): ",extz);
  minz = readlnint()-1; if (minz<0) minz = 0; if (minz>=extz) minz = extz-1;
  printf ("pindown> Maximum z value (%d-%d): ",minz+1,extz);
  maxz = readlnint()-1; if (maxz<minz) maxz = minz; if (maxz>=extz) maxz = extz-1;

  gridx += width*minx;
  gridy += width*miny;
  gridz += width*minz;

  nvox = (maxx-minx+1)*(maxy-miny+1)*(maxz-minz+1);
  phi2 = (MYFLT *) malloc(nvox * sizeof(MYFLT)); 
  if (phi2 == NULL) {
    fprintf(stderr, "pindown> Error: Could not satisfy memory allocation request [e.c. 32010]\n"); 
    exit(32010);
  }

  count = 0;
  for(indz=minz;indz<=maxz;indz++) 
  for(indy=miny;indy<=maxy;indy++) 
  for(indx=minx;indx<=maxx;indx++) {
    *(phi2+count) = *(phi+pindex(indx,indy,indz));
    ++ count;
  }    
  free (phi);
  writevol(argv[2], width, gridx, gridy, gridz, maxx-minx+1, maxy-miny+1, maxz-minz+1, phi2);
}

unsigned long pindex(int i, int j, int k) 
{ return extx*exty*k+extx*j+i; }
